package propellerbug;

import gnu.io.CommPort;
import gnu.io.CommPortIdentifier;
import gnu.io.NoSuchPortException;
import gnu.io.SerialPort;
import gnu.io.UnsupportedCommOperationException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author matthew stevenson
 */
public class PropellerServoController {
    private final static Logger theLogger = Logger.getLogger(PropellerServoController.class.getName());
    private SerialPort myPort;
    private OutputStream myPortWriter;
    private InputStream myPortReader;
    private String myPortName;

    public PropellerServoController(String portName){
        myPortName = portName;
    }
    
    public void connect() throws Throwable{
        CommPortIdentifier portIdentifier;
        try{
            portIdentifier = CommPortIdentifier.getPortIdentifier(myPortName);
        }catch(NoSuchPortException ex){
            throw new IllegalArgumentException("Port (" + myPortName + ") not found.", ex);
        }
        if(portIdentifier.isCurrentlyOwned()){
            throw new IllegalAccessError("Port (" + myPortName + ") in use by another program or process and cannot be accessed.");
        }
        CommPort commPort = portIdentifier.open(this.getClass().getName(),2000);
        if(!(commPort instanceof SerialPort)){
            throw new IllegalArgumentException("Port: " + myPortName + " is not a Serial Port.");
        }
        myPort = (SerialPort)commPort;
        myPort.setSerialPortParams(2400,SerialPort.DATABITS_8,SerialPort.STOPBITS_1,SerialPort.PARITY_NONE);
        myPortReader = myPort.getInputStream();
        myPortWriter = myPort.getOutputStream();
        Utils.sleep(2000);
    }

    public void disconnect() {
        try{
            myPortReader.close();
            myPortWriter.close();
        }catch(IOException ex){
            theLogger.log(Level.WARNING, "Unable to close Serial port IO Streams.", ex);
        }
        myPort.close();
    }

    public boolean moveServo(byte id, byte speed, byte low, byte high){
        int pos = Integer.rotateLeft(Utils.unsign(high), 8) + Utils.unsign(low);
        System.out.println("Moving Servo " + id + " to " + pos + " (High: " + Utils.unsign(high) + ", Low: " + Utils.unsign(low) + ")");
        if(!write(id, speed, low, high)){
            theLogger.log(Level.SEVERE, "Cannot move servo {0}, unable to write to serial port {1}", new Object[]{id, myPortName});
            return false;
        }
        return true;
    }

    public String getVersion() throws IOException{
        flushReader();
        if(!write('V','E','R','?')){
            throw new IOException("Cannot request version, unable to write to serial port " + myPortName);
        }
        Utils.sleep(75);   //need to give the board some time to respond.
        try{
            return Utils.bytesToString(read());
        }catch(IOException ex){
            throw new IOException("There was a problem reading version number from serial port " + myPortName, ex);
        }
    }

    public void setBaudRate(int rate) throws IOException{
        flushReader();
        int br = rate == 2400 ? 0 : 1;
        if(!write('S','B','R', br)){
            throw new IOException("Cannot set baud rate, unable to write to serial port " + myPortName);
        }
        Utils.sleep(75);   //need to give the board some time to respond.
        try{
            myPort.setSerialPortParams(rate,SerialPort.DATABITS_8,SerialPort.STOPBITS_1,SerialPort.PARITY_NONE);
            getVersion();
        }catch(IOException ex){
            throw new IOException("Unable to read baud rate change reply from serial port " + myPortName, ex);
        }catch(UnsupportedCommOperationException ex){
            throw new IOException("Unable to change baudrate for serial port " + myPortName, ex);
        }
    }

    public byte[] getPosition(int id) throws IOException{
        flushReader();
        if(!write('R','S','P', id)){
            throw new IOException("Unable to read servo position for ID " + id + ".  Error writing request to serial port " + myPortName);
        }
        Utils.sleep(75);   //need to give the board some time to respond.
        try{
            byte[] bytes = read();

            System.out.println("Number of Bytes Read: " + bytes.length);
            if(bytes.length >= 3){
                int pos = Integer.rotateLeft(Utils.unsign(bytes[1]), 8) + Utils.unsign(bytes[2]);
                System.out.println("Read Servo " + bytes[0] +
                        ". Pos: " + pos + ", High: " + Utils.unsign(bytes[1]) +
                        ", Low: " + Utils.unsign(bytes[2]));
            }
            return bytes;
        }catch(IOException ex){
           throw new IOException("Unable to read servo position for ID " + id + ".  Error reading serial port " + myPortName, ex);
        }
    }

    private boolean write(Object... data){
        byte[] bytes = new byte[data.length + 4];
        bytes[0] = '!';
        bytes[1] = 'S';
        bytes[2] = 'C';
        for(int i=0; i<data.length; i++){
            if(data[i] instanceof Integer){
                bytes[i+3] = ((Integer)data[i]).byteValue();
            }else if(data[i] instanceof Character){
                bytes[i+3] = (byte)(char)((Character)data[i]);
            }else if(data[i] instanceof Byte){
                bytes[i+3] = ((Byte)data[i]);
            }else {
                theLogger.log(Level.SEVERE, "Cannot convert {0} to Byte.  Unable to write data to serial port.", data[i].getClass().getName());
                return false;
            }
        }
        bytes[data.length+3] = 13;
        try{
            myPortWriter.write(bytes);
            myPortWriter.flush();
        }catch(IOException ex){
            theLogger.log(Level.SEVERE, "Unable to write to serial port " + myPortName, ex);
            return false;
        }
        return true;
    }

    private void flushReader(){
        try{
            int n = myPortReader.available();
            for(int i=0; i<n; i++){myPortReader.read();}
        }catch(IOException ex){
            theLogger.log(Level.WARNING, "There was a problem flush the reader for serial port " + myPortName, ex);
        }
    }

    private byte[] read() throws IOException{
        int n = myPortReader.available();
        byte[] read = new byte[n];
        myPortReader.read(read);
        if(read.length != 3){
            theLogger.log(Level.WARNING, "Error reading serial port {0}.  Expected 3 bytes, found {1}",
                    new Object[]{myPortName, read.length});
            Utils.sleep(25); //wait to print warning message
        }
        return read;
    }
}
